const {when: unload} = require('sdk/system/unload');
const {getPref, setPref} = require('prefs');
const {EventEmitter} = require('utils/event_emitter');

let browsec = function() {
  let proxyService = require('proxy');
  let analytics = require('analytics');
  let instrumentation = require('instrumentation');
  let utils = require('utils');

  let log = require('logging').create("browsec.js", "Debug");

  let localProxyHost = "127.0.0.1";
  let localProxyPort = 49736;

  // instrumentation state
  let instrumenter = null;
  let enabled = false;
  let status = null;

  let emitter = EventEmitter();

  return {
    init : function() {
      try {
        log.debug("browsec.init()");
      } catch (e) {
        log.error("Error in init: " + e);
      }
    },

    on: function(event, callback) {
      emitter.on(event, callback);
    },

    off: function(event, callback) {
      emitter.off(event, callback);
    },

    get enabled() {
      return enabled;
    },

    set enabled(value) {
      enabled = value;
      emitter.emit('enabled', value);
    },

    get status() {
      return status;
    },

    set status(value) {
      status = value;
      emitter.emit('status', value);
    },

    get active() {
      return getPref('active');
    },

    set active(newActiveState) {
      setPref("active", newActiveState);
      emitter.emit('active', newActiveState);
    },

    start : function() {
      try {
        const remotePort = 443;
        const remoteHost = 'postlm.com';

        log.info("starting tunnel");
        let tunnel = require('./tunnel').createServer(remoteHost, remotePort).listen(localProxyPort);
        unload(function() {
          // stop listening; active connection will continue operating
          tunnel.close();
        });
      } catch (e) {
        log.error("Error in start: " + e);
      }
    },

    install_addon : function() {
      try {
        log.info("Start installing addon");

        let localProxyUrl = this.getLocalProxyUrl();
        let active = this.active;
        if (active) {
          proxyService.setProxy(localProxyUrl);
        }
      } catch (e) {
        log.error("Error installing addon: " + e);
      }
    },

    getLocalProxyUrl : function() {
      return localProxyHost + ":" + localProxyPort;
    },

    /**
     * Ensures that proxy settings correspond to addon's `active` state.
     *
     * {Active, _} -> {Active, ProxySet}
     */
    enable_addon : function() {
      try {
        log.debug("Enable addon");

        if (this.active) {
          this.status = 'active';
          this.activate();
        } else {
          this.status = 'inactive';
          // TODO: deactivate?()
        }
        unload(this.disable_addon.bind(this));

      } catch (e) {
        log.error("enable_addon error: " + e);
      }
    },

    /**
     * Ensure that proxy settings are cleared when addon is unloaded
     *
     * {_, _} -> {_, ProxyNotSet}
     */
    disable_addon : function () {
      try {
        log.debug("Disable addon");

        if (this.active) {
          this.deactivate(true);
        }

      } catch (e) {
        log.error("disable_addon error: " + e);
      }      
    }, 

    toggleActiveState: function() {
      try {
        log.debug("clickToolbarButton");

        let active = this.active;
        let newActiveState = !active;

        this.active = newActiveState;
        if (newActiveState) {
          browsec.activate();
        } else {
          browsec.deactivate();
        }
        return newActiveState;
      } catch (err) {
        log.error("Error in clickToolbarButton: " + err);
      }
    },

    activate : function() {
      try {
        log.debug("activate");
        this.status = 'active';

        let localProxyUrl = this.getLocalProxyUrl();

        let bypass = getPref('bypass');

        analytics.trackEvent("activate");

        instrumenter = instrumentation.start();

        if (!proxyService.isProxySet(localProxyUrl)) {
          proxyService.backupBrowserProxySettings();
        }

        proxyService.setProxy(localProxyUrl, bypass);
      } catch (err) {
        log.error("Error in activate: " + err);
      }
    },

    deactivate : function(onShutdown) {
      if (typeof onShutdown === 'undefined') { onShutdown = false; }

      try {
        log.debug("deactivate");
        if (!onShutdown) {
          this.status = 'inactive';
        }

        // Save current bypass string to extension settings
        setPref("bypass", proxyService.getBypass());

        proxyService.restoreBrowserProxySettings();
        proxyService.clearProxySettingsBackup();

        if (!onShutdown) {
          analytics.trackEvent("deactivate");
        }

        if (!onShutdown) {
          if (instrumenter) {
            instrumentation.stop(instrumenter);
            instrumenter = null;
          }
        }
      } catch (err) {
        log.error("Error in deactivate: " + err);
      }
    }, 

    getLocalProxyHost : function() {
      return localProxyHost;
    },

    getLocalProxyPort : function() {
      return localProxyPort;
    }
  };
}();

module.exports = browsec;
