const {Cc, Ci} = require('chrome');
var analytics = require('analytics');
let log = require('logging').create('instrumentation.js', 'Debug');
const {when: unload} = require('sdk/system/unload');

const SAMPLE_RATE = 1000;

/**
 * An nsIObserver implementation that reports each `SAMPLE_RATE`th event to GA.
 */
function TrackingObserver() {
  // a request counter
  this.count = 0;

  // when true, suspends tracking by this observer
  //
  // suspension relies on the assumption that when issuing XHR requests,
  // observer will be notified before the call to xhr.send() returns.
  this.suspended = false;
}

TrackingObserver.prototype = {
  observe: function(subject, topic, data) {
    if(this.suspended) {
      return;
    }

    this.count = (this.count + 1) % SAMPLE_RATE;
    if (this.count == 0) {
      // suspend observer while tracking events, otherwise it will get notified
      // of HTTP requests to GA, which would trigger an infinite loop.
      this.suspended = true;
      try {
        analytics.trackEvent('activity');
      } catch(e) {
        log.error(e);
      } finally {
        this.suspended = false;
      }
    }
  }
};

/**
 * Currently active observers. Should be either empty or a single-element list.
 */
let observers = [];

/**
 * Starts reporting every SAMPLE_RATE request to the analytics
 *
 * @return handler that can be used passed to stop() to stop tracking.
 */
function start() {
  let observerService = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
  let observer = new TrackingObserver();
  observerService.addObserver(observer, "http-on-opening-request", false);
  observers.push(observer);
  return observer;
}

/**
 * Stops tracking user activity.
 *
 * @param observer nsIObserver observer returned from `start()`.
 */
function stop(observer) {
  console.log("Stopping obsever: " + observer);
  let observerService = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
  observerService.removeObserver(observer, "http-on-opening-request");
  let index = observers.indexOf(observer);
  if (index >= 0) {
    observers.splice(index, 1);
  }
}

unload(function() {
  let observerService = Cc["@mozilla.org/observer-service;1"].getService(Ci.nsIObserverService);
  for (let observer of observers) {
    try {
      observerService.removeObserver(observer, "http-on-opening-request");
    } catch (e) {
      log.error(e);
    }
  }
  observers = null;
});

exports.start = start;
exports.stop = stop;
